from ipywidgets import *  # noqa


NUMERIC_CONTROL_WIDGETS = (
    IntSlider,
    # FloatSlider,  # floats suck
    # FloatLogSlider,  # floats suck
    IntRangeSlider,
    # FloatRangeSlider,  # floats suck
    BoundedIntText,
    # BoundedFloatText,  # floats suck
    # IntText,  # No open ended
    # FloatText,  # No open ended
)
NUMERIC_OUTPUT_WIDGETS = NUMERIC_CONTROL_WIDGETS + (IntProgress, FloatProgress)

BOOLEAN_CONTROL_WIDGETS = (ToggleButton, Checkbox)
BOOLEAN_OUTPUT_WIDGETS = BOOLEAN_CONTROL_WIDGETS + (Valid,)

SELECTION_CONTROL_WIDGETS = (
    Dropdown,
    RadioButtons,
    Select,
    SelectionSlider,
    ToggleButtons,
    SelectionRangeSlider,
    SelectMultiple,
)
SELECTION_OUTPUT_WIDGETS = SELECTION_CONTROL_WIDGETS

STRING_CONTROL_WIDGETS = (
    # Text,  # No open ended
    # Textarea,  # No opeen ended
)
STRING_OUTPUT_WIDGETS = (
    Label,
    # HTML,  # TODO
    # HTMLMath,  # TODO
    # Image,  # TODO
)

OTHER_CONTROL_WIDGETS = (
    # Button,  # TODO
    # Play,  # TODO
    # DatePicker,  # TODO
    # ColorPicker  # TODO
)

CONTROL_WIDGETS = (
    NUMERIC_CONTROL_WIDGETS
    + BOOLEAN_CONTROL_WIDGETS
    + SELECTION_CONTROL_WIDGETS
    + STRING_CONTROL_WIDGETS
    + OTHER_CONTROL_WIDGETS
)

OUTPUT_WIDGETS = (
    NUMERIC_OUTPUT_WIDGETS
    + BOOLEAN_OUTPUT_WIDGETS
    + SELECTION_OUTPUT_WIDGETS
    + STRING_OUTPUT_WIDGETS
    + (Output,)
)

VALUE_WIDGETS = CONTROL_WIDGETS + OUTPUT_WIDGETS
LAYOUT_WIDGETS = Box, HBox, VBox, Accordion, Tab



import io
import csv
import copy
import json
import itertools

from nbconvert.preprocessors import Preprocessor

from illusionist import widgets, kernel_utils, utils
from illusionist.client import IllusionistClient
from illusionist.utils import DEV_MODE


NUMERIC_CONTROL_WIDGETS = (
    "IntSliderModel",
    # "FloatSlider",  # floats suck
    # "FloatLogSlider",  # floats suck
    "IntRangeSliderModel",
    # "FloatRangeSlider",  # floats suck
    "BoundedIntTextModel",
    # "BoundedFloatText",  # floats suck
    # "IntText",  # No open ended
    # "FloatText",  # No open ended
)
NUMERIC_OUTPUT_WIDGETS = NUMERIC_CONTROL_WIDGETS + (
    "IntProgressModel",
    "FloatProgressModel",
)

BOOLEAN_CONTROL_WIDGETS = ("ToggleButtonModel", "CheckboxModel")
BOOLEAN_OUTPUT_WIDGETS = BOOLEAN_CONTROL_WIDGETS + ("ValidModel",)

SELECTION_CONTROL_WIDGETS = (
    "DropdownModel",
    "RadioButtonsModel",
    "SelectModel",
    "SelectionSliderModel",
    "ToggleButtonsModel",
    "SelectionRangeSliderModel",
    "SelectMultipleModel",
)
SELECTION_OUTPUT_WIDGETS = SELECTION_CONTROL_WIDGETS

STRING_CONTROL_WIDGETS = (
    # "Text",  # No open ended
    # "Textarea",  # No opeen ended
)
STRING_OUTPUT_WIDGETS = (
    "LabelModel",
    # "HTMLModel",  # TODO
    # "HTMLMathModel",  # TODO
    # "ImageModel",  # TODO
)

OTHER_CONTROL_WIDGETS = (
    # "ButtonModel",  # TODO
    # "PlayModel",  # TODO
    # "DatePickerModel",  # TODO
    # "ColorPickerModel"  # TODO
)

CONTROL_WIDGETS = (
    NUMERIC_CONTROL_WIDGETS
    + BOOLEAN_CONTROL_WIDGETS
    + SELECTION_CONTROL_WIDGETS
    + STRING_CONTROL_WIDGETS
    + OTHER_CONTROL_WIDGETS
)

OUTPUT_WIDGETS = (
    NUMERIC_OUTPUT_WIDGETS
    + BOOLEAN_OUTPUT_WIDGETS
    + SELECTION_OUTPUT_WIDGETS
    + STRING_OUTPUT_WIDGETS
    + ("OutputModel",)
)

VALUE_WIDGETS = CONTROL_WIDGETS + OUTPUT_WIDGETS


WIDGET_ONCHANGE_MIMETYPE = "application/vnd.illusionist.widget-onchange+json"


class IllusionistPreprocessor(Preprocessor, IllusionistClient):
    """
    Execute cells in the notebook
    Then looks at the widgets and generates on-change values

    This class is mostly copied from nbconvert.preprocessors.execute
    """

    def __init__(self, **kw):
        nb = kw.get("nb")
        if nb:
            del kw["nb"]
        Preprocessor.__init__(self, nb=nb, **kw)
        IllusionistClient.__init__(self, nb, **kw)

    def preprocess(self, nb, resources=None, km=None):
        # self.log_level = "DEBUG"
        self.nb = nb
        self.km = km
        self.onchange_values = {}

        resources = resources if resources else {}
        resources["illusionist_devmode"] = DEV_MODE
        if resources["illusionist_devmode"]:
            self.log.warning(
                "Illusionist DevMode is ON. "
                "Output Notebooks and HTML will contain extra cells at the end"
            )

        try:
            self.reset_execution_trackers()
            self.execute(cleanup_kc=False)

            self.nb.metadata.widgets.update(
                {WIDGET_ONCHANGE_MIMETYPE: self.widget_onchange_state}
            )
        finally:
            # Clean up
            self._cleanup_kernel()

        return nb, resources

    def post_exec(self):
        """
        This methond it's executed as part for self.execute()
        """
        # Load helper code into the kernel
        _ = self.run_code(utils.get_source(widgets))
        _ = self.run_code(utils.get_source(kernel_utils))

        # Save Notebook and widget state before executing extra code
        # nb_cells_before = copy.deepcopy(self.nb.cells)
        widget_state_before = copy.deepcopy(self.widget_state)

        value_widgets = self.run_code_eval("get_widgets_ids(kind='value')")
        control_widgets = self.run_code_eval("get_widgets_ids(kind='control')")

        # 1. Iterate the control widgets and see which outputs it affects
        affected_by = {m_id: set() for m_id in value_widgets}

        for widget_id in control_widgets:
            init_state = copy.deepcopy(self.widget_state)
            possible_values = self.possible_values(widget_id)
            widget_affects = []

            for value in possible_values[:2]:
                self.run_code(f"set_widget_value('{widget_id}', {value})")

                new_state = self.widget_state
                diff = diff_state(init_state, new_state, my_id=widget_id)
                # print(diff)
                widget_affects.extend(diff)

            for affected in widget_affects:
                affected_by[affected] |= {widget_id}

        # 2. Iterate affected_by and add matrix (per output widget) to the matrix

        # matrices is of the form: {output_id: [[ ... matrix ... ]] }
        matrices = {}

        for output_widget_id, input_widget_ids in affected_by.items():
            if len(input_widget_ids) > 0:
                values = self.widget_matrix(output_widget_id, input_widget_ids)
                matrices[output_widget_id] = {
                    "affected_by": list(input_widget_ids),
                    "values": values,
                }

        # Save the onChange state
        onChangeState = {"version_major": 1, "version_minor": 0}
        onChangeState["all_widgets"] = value_widgets
        onChangeState["control_widgets"] = control_widgets
        onChangeState["onchange"] = matrices
        self.widget_onchange_state = onChangeState

        # Set the original widget_state values back
        self.widget_state = widget_state_before

    def widget_matrix(self, output_widget_id, input_widget_ids):
        output_state = self.widget_state[output_widget_id]
        input_states = [self.widget_state[w_id] for w_id in input_widget_ids]

        # Get the product of all possible input values
        product = widget_product(input_states)

        # 2. Now we iterate the combinations of possible values
        # To create the matrix

        matrix = {}
        outputs = []
        input_ids = list()
        for inputs_set in product:

            # Update values of input widgets
            for widget_id, value in zip(input_widget_ids, inputs_set):
                self.run_code(f"set_widget_value('{widget_id}', {value})")

            # Save the new value of the output widget
            hash_ = self.hash_fn(input_widget_ids)
            matrix[hash_] = self.get_output_value(output_widget_id)

        return matrix

    def get_output_value(self, output_widget_id):
        widget_state = self.widget_state[output_widget_id]
        model_name = widget_state["_model_name"]

        if model_name in NUMERIC_OUTPUT_WIDGETS:
            return widget_state["value"]
        elif model_name in BOOLEAN_OUTPUT_WIDGETS:
            return widget_state["value"]
        elif model_name in SELECTION_OUTPUT_WIDGETS:
            if isinstance(widget_state["index"], tuple):
                return list(widget_state["index"])
            return widget_state["index"]
        elif model_name in STRING_OUTPUT_WIDGETS:
            return widget_state["value"]
        elif model_name == "OutputModel":
            return widget_state["outputs"]
        else:
            raise Exception(f"Output Widget type '{model_name}' not supported.")

    def hash_fn(self, widget_ids):
        widget_states = []
        # print(self.widget_state)
        for widget_id in widget_ids:
            w_state = self.widget_state[widget_id]
            widget_states.append(w_state)

        return hash_fn(widget_states)

    def possible_values(self, widget_id):
        """
        Returns a list with the possible values for a widget
        """
        widget_state = self.widget_state[widget_id]
        return possible_values(widget_state)


def widget_product(input_states):
    # Make a product of all the possible widget values
    # For each input_widgets, get all possible values they can have
    possible_values_by_widget = []
    for w_state in input_states:
        values = possible_values(w_state)
        possible_values_by_widget.append(values)

    return list(itertools.product(*possible_values_by_widget))


def diff_state(initial, new, my_id=None):
    """
    Return a list of widget model_ids that have changed based on the two widget states
    """
    diff = []
    for (init_id, init_state), (new_id, new_state) in zip(initial.items(), new.items()):
        assert init_id == new_id
        if init_state != new_state:
            diff.append(init_id)

    if my_id and my_id in diff:
        diff.remove(my_id)
    return diff


def possible_values(widget_state):
    model_name = widget_state["_model_name"]
    if model_name == "IntRangeSliderModel":
        # Return all combinations that are possible: low < high
        range_ = range(
            widget_state["min"],
            widget_state["max"] + widget_state["step"],
            widget_state["step"],
        )
        ret = itertools.product(range_, range_)
        ret = [[i, j] for i, j in ret if i <= j]
        return ret
    elif model_name in NUMERIC_CONTROL_WIDGETS:
        return list(
            range(
                widget_state["min"],
                widget_state["max"] + widget_state["step"],
                widget_state["step"],
            )
        )
    elif model_name in BOOLEAN_CONTROL_WIDGETS:
        return [True, False]
    elif model_name == "SelectionRangeSliderModel":
        range_ = range(0, len(widget_state["_options_labels"]))
        ret = itertools.product(range_, range_)
        ret = [[i, j] for i, j in ret if i <= j]
        return ret
    elif model_name == "SelectMultipleModel":
        range_ = range(0, len(widget_state["_options_labels"]))
        return list(powerset(range_))
    elif model_name in SELECTION_CONTROL_WIDGETS:
        range_ = range(0, len(widget_state["_options_labels"]))
        return list(range_)
    else:
        raise Exception(f"Widget type '{model_name}' not supported.")


def powerset(iterable):
    """
    Example: powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)
    """
    s = list(iterable)
    return itertools.chain.from_iterable(
        itertools.combinations(s, r) for r in range(len(s) + 1)
    )


def hash_fn(w_states):
    values = []
    for w_state in w_states:
        model_name = w_state["_model_name"]

        if model_name in NUMERIC_CONTROL_WIDGETS:
            value = w_state["value"]
        elif model_name in BOOLEAN_CONTROL_WIDGETS:
            value = w_state["value"]
        elif model_name in SELECTION_CONTROL_WIDGETS:
            value = w_state["index"]
        else:
            raise Exception(f"Cannot use '{model_name}' as input")

        if isinstance(value, str):
            value = json.dumps(value)
        elif isinstance(value, bool):
            value = json.dumps(value)
        elif isinstance(value, (tuple, list)):
            value = list(value)
            value = json.dumps(value, separators=(",", ":"))
        values.append(value)

    return dumps(values)


def dumps(values):
    """
    CSV Serialization
    """
    output = io.StringIO()
    writer = csv.writer(
        output,
        skipinitialspace=True,
        delimiter=",",
        quotechar='"',
        quoting=csv.QUOTE_NONNUMERIC,
    )
    writer.writerow(values)
    contents = output.getvalue()
    output.close()
    return contents.strip()


if __name__ == "__main__":
    from nbconvert.nbconvertapp import main

    main()



import os
import sys
import os.path

import jinja2
from nbconvert.exporters.html import HTMLExporter
from traitlets import List, default

from illusionist.preprocessor import IllusionistPreprocessor


@jinja2.contextfunction
def include_template(ctx, name):
    """Include a file relative to this python file"""
    env = ctx.environment
    return jinja2.Markup(env.loader.get_source(env, name)[0])


@jinja2.contextfunction
def include_external_file(ctx, name):
    """Include an encoded base64 image"""
    with open(os.path.abspath(name), "r") as f:
        content = f.read()
    return jinja2.Markup(content)


@jinja2.contextfunction
def include_external_base64_img(ctx, name):
    """Include an encoded base64 external image"""
    import base64

    with open(os.path.abspath(name), "rb") as f:
        encoded_string = base64.b64encode(f.read())
    return jinja2.Markup(encoded_string.decode())


class IllusionistHTMLExporter(HTMLExporter):
    # Name for the menu item under "File -> Download as" in the IDE
    export_from_notebook = "Illusionist HTML"
    preprocessors = [IllusionistPreprocessor]

    @property
    def template_path(self):
        """
        Append template intalled to share
        This is compat code until nbconvert 6.0.0 lands
        The structure of the project here is whats 6.0.0 will use
        """
        return super().template_path + [
            os.path.join(
                sys.prefix, "share", "jupyter", "nbconvert", "templates", "illusionist"
            )
        ]

    def _template_file_default(self):
        """
        We want to use the new template we ship with our library.
        """
        return "index"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        globals_ = {}
        globals_["include_template"] = include_template
        globals_["include_external_file"] = include_external_file
        globals_["include_external_base64_img"] = include_external_base64_img
        self.environment.globals.update(globals_)

    def default_filters(self):
        for pair in super().default_filters():
            yield pair
        yield ("test_filter", self.test_filter)

    def test_filter(self, text):
        return "test_filter: " + text



import pytest

from illusionist.preprocessor import hash_fn
from illusionist.widgets import *  # noqa

from .utils import preprocessor


@pytest.mark.parametrize("class_", [IntSlider, BoundedIntText])
def test_hash_fn_int_single_value(class_):
    w = class_(min=1, max=5)
    w_states = [w.get_state() for w in [w]]
    assert hash_fn(w_states) == "1"

    w.value = 5
    w_states = [w.get_state() for w in [w]]
    assert hash_fn(w_states) == "5"


@pytest.mark.parametrize("class_", [IntSlider, BoundedIntText])
def test_hash_fn_int_multiple_single_value(class_):
    w1 = class_(min=1, max=5)
    w2 = class_(min=6, max=10)
    w_states = [w.get_state() for w in [w1, w2]]
    assert hash_fn(w_states) == "1,6"

    w1.value = 5
    w2.value = 8
    w_states = [w.get_state() for w in [w1, w2]]
    assert hash_fn(w_states) == "5,8"


@pytest.mark.parametrize("class_", BOOLEAN_CONTROL_WIDGETS)
def test_hash_fn_boolean(class_):
    w = class_()
    w_states = [w.get_state() for w in [w]]
    assert hash_fn(w_states) == '"false"'

    w.value = True
    w_states = [w.get_state() for w in [w]]
    assert hash_fn(w_states) == '"true"'


@pytest.mark.parametrize("class_", BOOLEAN_CONTROL_WIDGETS)
def test_hash_fn_boolean_multiple(class_):
    w1 = class_()
    w2 = class_()
    w_states = [w.get_state() for w in [w1, w2]]
    assert hash_fn(w_states) == '"false","false"'

    w1.value = True
    w_states = [w.get_state() for w in [w1, w2]]
    assert hash_fn(w_states) == '"true","false"'

    w2.value = True
    w_states = [w.get_state() for w in [w1, w2]]
    assert hash_fn(w_states) == '"true","true"'


@pytest.mark.parametrize(
    "class_", [SelectMultiple],
)
def test_hash_fn_selection_multi_value(class_):
    options = ["a", "b"]
    w = class_(options=options)
    w_states = [w.get_state() for w in [w]]
    assert hash_fn(w_states) == '"[]"'

    w.value = ["a"]
    w_states = [w.get_state() for w in [w]]
    assert hash_fn(w_states) == '"[0]"'

    w.value = ["b"]
    w_states = [w.get_state() for w in [w]]
    assert hash_fn(w_states) == '"[1]"'

    w.value = ["a", "b"]
    w_states = [w.get_state() for w in [w]]
    assert hash_fn(w_states) == '"[0,1]"'



import ipywidgets
import pytest

from illusionist.preprocessor import possible_values
from illusionist.widgets import *  # noqa

from .utils import preprocessor


def test_widget_matrix_IntSlider(preprocessor):
    preprocessor.run_code("w = IntSlider(min=-3, max=0)")
    preprocessor.run_code("lbl = Label()")
    in_id = preprocessor.run_code_eval("w.model_id")
    out_id = preprocessor.run_code_eval("lbl.model_id")
    in_ids = [in_id]

    matrix = preprocessor.widget_matrix(out_id, in_ids)
    assert list(matrix.keys()) == ["-3", "-2", "-1", "0"]


def test_widget_matrix_IntRangeSlider(preprocessor):
    preprocessor.run_code("w = IntRangeSlider(min=0, max=2)")
    preprocessor.run_code("lbl = Label()")
    in_id = preprocessor.run_code_eval("w.model_id")
    out_id = preprocessor.run_code_eval("lbl.model_id")
    in_ids = [in_id]

    matrix = preprocessor.widget_matrix(out_id, in_ids)
    ans = ['"[0,0]"', '"[0,1]"', '"[0,2]"', '"[1,1]"', '"[1,2]"', '"[2,2]"']
    assert list(matrix.keys()) == ans


@pytest.mark.parametrize(
    "class_",
    [
        # "Dropdown",
        # "RadioButtons",
        # "Select",
        # "SelectionSlider",
        # "ToggleButtons",
        "SelectMultiple",
    ],
)
def test_widget_matrix_Selection(preprocessor, class_):
    preprocessor.run_code(f'w = {class_}(options=["a", "b", "c"])')
    preprocessor.run_code("lbl = Label()")
    in_id = preprocessor.run_code_eval("w.model_id")
    out_id = preprocessor.run_code_eval("lbl.model_id")
    in_ids = [in_id]

    matrix = preprocessor.widget_matrix(out_id, in_ids)
    ans = [
        '"[]"',
        '"[0]"',
        '"[1]"',
        '"[2]"',
        '"[0,1]"',
        '"[0,2]"',
        '"[1,2]"',
        '"[0,1,2]"',
    ]
    assert list(matrix.keys()) == ans



import os
import sys
import shutil

from setuptools import find_packages, setup
from setuptools.command.develop import develop


setup_dir = os.path.abspath(os.path.dirname(__file__))


def read_file(filename):
    filepath = os.path.join(setup_dir, filename)
    with open(filepath) as file:
        return file.read()


scm_version_write_to_prefix = os.environ.get(
    "SETUPTOOLS_SCM_VERSION_WRITE_TO_PREFIX", setup_dir
)


def parse_git(root, **kwargs):
    from setuptools_scm.git import parse

    kwargs["describe_command"] = 'git describe --dirty --tags --long --match "[0-9].*"'
    return parse(root, **kwargs)


def get_data_files():
    # Add the templates
    data_files = []
    for (dirpath, dirnames, filenames) in os.walk("share/jupyter/nbconvert/templates/"):
        if filenames:
            data_files.append(
                (dirpath, [os.path.join(dirpath, filename) for filename in filenames])
            )
    return data_files


class DevelopCmd(develop):
    """The DevelopCmd will create symlinks for nbconvert under:
    sys.prefix/share/jupyter/
    """

    prefix_targets = [
        ("share/jupyter/nbconvert/templates", "illusionist"),
    ]

    def run(self):
        for parent, name in self.prefix_targets:
            source = os.path.join(os.path.abspath(parent), name)
            target = os.path.join(sys.prefix, parent, name)
            target = target.rstrip(os.path.sep)

            if not os.path.exists(os.path.dirname(target)):
                print("Creating:", os.path.dirname(target))
                os.makedirs(os.path.dirname(target))

            if os.path.islink(target):
                print("Removing link:", target)
                os.remove(target)

            print("Linking", source, "->", target)
            os.symlink(source, target)

        super(DevelopCmd, self).run()


setup(
    name="illusionist",
    use_scm_version={
        "root": os.path.dirname(setup_dir),
        "parse": parse_git,
        "write_to": os.path.join(
            scm_version_write_to_prefix, "illusionist/_generated_version.py"
        ),
    },
    packages=find_packages(),
    # package_dir={"": "src"},
    zip_safe=False,
    include_package_data=True,
    # package_data={"illusionist": ["includes/*"]},
    data_files=get_data_files(),
    cmdclass={"develop": DevelopCmd},
    entry_points={
        "nbconvert.exporters": [
            "illusionist = illusionist:IllusionistHTMLExporter",
            "illusionist-nb = illusionist:IllusionistNotebookExporter",
        ],
    },
    options={"bdist_wheel": {"universal": "1"}},
    python_requires=">=3.6",
    setup_requires=["setuptools_scm"],
    install_requires=read_file("requirements-package.txt").splitlines(),
    extras_require={
        "test": ["pytest"],
        "dev": read_file("requirements.txt").splitlines(),
    },
    description="",
    long_description=read_file("README.md"),
    long_description_content_type="text/markdown",
    license="Apache License, Version 2.0",
    maintainer="Daniel Rodriguez",
    maintainer_email="daniel@danielfrg.com",
    url="https://github.com/danielfrg/illusionist",
    keywords=[],
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



import os
import pkg_resources
from pip._internal.utils.misc import dist_is_editable

distributions = {v.key: v for v in pkg_resources.working_set}
DEV_MODE = dist_is_editable(distributions["illusionist"])
_ = os.environ.get("ILLUSIONIST_DEV_MODE")
DEV_MODE = False if _ == "0" else DEV_MODE
# DEV_MODE = False
del distributions


def get_source(code):
    import inspect

    return inspect.getsource(code)



import ipywidgets
import pytest

from .utils import preprocessor


# def test_pb(preprocessor):
# # def test_pb(preprocessor):
#     # preprocessor.run_code("import ipywidgets")
#     # preprocessor.run_code("w = ipywidgets.IntSlider(min=5, max=10)")
#     # w_id = preprocessor.run_code_eval("w.model_id")

#     w = ipywidgets.IntSlider(min=5, max=10)
#     w_state = w.get_state()
#     # w_id = list(preprocessor.widget_state.keys())[0]
#     assert possible_values(w_state)
#     # print(preprocessor.widget_state)

#     preprocessor._cleanup_kernel()



import ipywidgets
import pytest

from illusionist.preprocessor import possible_values
from illusionist.widgets import *  # noqa


@pytest.mark.parametrize("class_", [IntSlider, BoundedIntText])
def test_possible_values_int_single_value(class_):
    w = class_(min=1, max=5)
    w_state = w.get_state()
    assert possible_values(w_state) == [1, 2, 3, 4, 5]

    w = class_(min=-3, max=2)
    w_state = w.get_state()
    assert possible_values(w_state) == [-3, -2, -1, 0, 1, 2]


def test_possible_values_IntRangeSlider():
    w = IntRangeSlider(min=0, max=3)
    w_state = w.get_state()
    a = [[0, 0], [0, 1], [0, 2], [0, 3], [1, 1], [1, 2], [1, 3], [2, 2], [2, 3], [3, 3]]
    assert possible_values(w_state) == a


@pytest.mark.parametrize("class_", [ToggleButton, Checkbox])
def test_possible_values_boolean(class_):
    w = class_()
    w_state = w.get_state()
    assert possible_values(w_state) == [True, False]


@pytest.mark.parametrize(
    "class_", [Dropdown, RadioButtons, Select, SelectionSlider, ToggleButtons,],
)
def test_possible_values_selection_single_value(class_):
    options = ["a", "b", "c"]
    w = class_(options=options)
    w_state = w.get_state()
    assert possible_values(w_state) == [0, 1, 2]

    options = [("One", 1), ("Two", 2), ("Three", 3)]
    w = class_(options=options)
    w_state = w.get_state()
    assert possible_values(w_state) == [0, 1, 2]


def test_possible_values_SelectMultiple():
    options = ["a"]
    w = SelectMultiple(options=options)
    w_state = w.get_state()
    assert possible_values(w_state) == [(), (0,)]

    options = ["a", "b"]
    w = SelectMultiple(options=options)
    w_state = w.get_state()
    assert possible_values(w_state) == [(), (0,), (1,), (0, 1)]

    options = ["a", "b", "c"]
    w = SelectMultiple(options=options)
    w_state = w.get_state()
    assert possible_values(w_state) == [
        (),
        (0,),
        (1,),
        (2,),
        (0, 1),
        (0, 2),
        (1, 2),
        (0, 1, 2),
    ]


def test_possible_values_SelectionRangeSlider():
    options = ["a", "b", "c"]
    w = SelectionRangeSlider(options=options)
    w_state = w.get_state()
    a = [[0, 0], [0, 1], [0, 2], [1, 1], [1, 2], [2, 2]]
    assert possible_values(w_state) == a


# def test_pb(preprocessor):
# # def test_pb(preprocessor):
#     # preprocessor.run_code("import ipywidgets")
#     # preprocessor.run_code("w = IntSlider(min=5, max=10)")
#     # w_id = preprocessor.run_code_eval("w.model_id")

#     w = IntSlider(min=5, max=10)
#     w_state = w.get_state()
#     # w_id = list(preprocessor.widget_state.keys())[0]
#     assert possible_values(w_state)
#     # print(preprocessor.widget_state)

#     preprocessor._cleanup_kernel()



import nbformat
from nbclient import NotebookClient
from nbclient.util import ensure_async, run_sync
from traitlets.config.application import Application

from illusionist.utils import DEV_MODE


class IllusionistClient(NotebookClient, Application):
    """
    Extends NBClient
    Adds utilities to run commands in the kernel
    """

    store_widget_state = True

    async def async_execute(self, reset_kc=False, **kwargs):
        """
        We copy and add a coupel of lines this method from nbclient.Client

        We need that extra code here because output widgets are
        not being updated if I run code in a function after even
        after keeping the kernel alive. Not sure why.
        """
        if reset_kc and self.km:
            await self._async_cleanup_kernel()
        self.reset_execution_trackers()

        async with self.async_setup_kernel(**kwargs):
            self.log.info("Executing notebook with kernel: %s" % self.kernel_name)
            for index, cell in enumerate(self.nb.cells):
                # Ignore `'execution_count' in content` as it's always 1
                # when store_history is False
                await self.async_execute_cell(
                    cell, index, execution_count=self.code_cells_executed + 1
                )

            # added
            await ensure_async(self.post_exec())
            # end add

            msg_id = await ensure_async(self.kc.kernel_info())
            info_msg = await self.async_wait_for_reply(msg_id)
            self.nb.metadata["language_info"] = info_msg["content"]["language_info"]
            self.set_widgets_metadata()

        return self.nb

    execute = run_sync(async_execute)

    def post_exec(self):
        """
        Overwrite this to execute extra code after the notebook has been executed
        """

    async def async_run_code(self, source, write_cell=None):
        write_cell = write_cell if write_cell is not None else DEV_MODE
        cell = nbformat.NotebookNode()
        cell.cell_type = "code"
        cell.execution_count = self.code_cells_executed + 1
        cell.metadata = {}
        cell.outputs = []
        cell.source = source

        self.nb["cells"].append(cell)
        cell_index = len(self.nb["cells"]) - 1
        # print(cell)
        cell = await self.async_execute_cell(
            cell, cell_index, execution_count=cell.execution_count
        )

        if write_cell == False:
            # Delete created cell
            del self.nb["cells"][cell_index]

        return cell

    run_code = run_sync(async_run_code)

    def run_code_eval(self, source):
        output = self.run_code(source)
        text_plain = output["outputs"][0]["data"]["text/plain"]
        return eval(text_plain)
